#include "stdafx.h"
#include "odbcgio.h" 

// Read the notes at the top of the ODBCGIO.H file

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

// Include the libraries to link
#ifdef WIN32
#pragma comment(lib, "odbc32.lib")
#else
#pragma comment(lib, "odbc.lib") // Does this pragma work in Win16?
#endif
 
#ifndef WIN32
#pragma message("Attention: Using ODBC requires a rather big stack otherwise errors can occur!")
#endif

//******************************************************************************
// ODBC Classes by G. Bavestrelli             Copyright Techint S.p.A. 1995-1999
// Any feedback is welcome, you can contact me at giovanni.bavestrelli@pomini.it
//******************************************************************************


// This function is declared but not defined
// You must define it in your application to handle errors
// It will be called with all the error information
BOOL ReportODBCProblem(BOOL isError,LPCSTR function,LPCSTR error,LPCSTR state,LPCSTR retcode,LPCSTR lastcommand,LPCSTR extra,long native,long row);


// Returns true if result code indicates success          
static inline BOOL Successful(RETCODE ret) 
{
  return ret==SQL_SUCCESS || ret==SQL_SUCCESS_WITH_INFO;
}

// Return true if the result code indicates a warning
static inline BOOL IsWarning(RETCODE ret)
{
  return ret==SQL_SUCCESS_WITH_INFO || ret==SQL_NO_DATA_FOUND;
}

// Return TRUE if result code indicates an error
static inline BOOL IsError(RETCODE ret)
{
  return !Successful(ret) && !IsWarning(ret); // ret==SQL_ERROR || ret==SQL_INVALID_HANDLE;
}

// String description of ODBC return code
static LPCSTR GetRetcodeDescription(RETCODE retcode)
{ 
  switch(retcode)
  { case SQL_INVALID_HANDLE     : return "SQL_INVALID_HANDLE";  
    case SQL_ERROR              : return "SQL_ERROR"         ;  
    case SQL_SUCCESS            : return "SQL_SUCCESS"       ;  
    case SQL_SUCCESS_WITH_INFO  : return "SQL_SUCCESS_WITH_INFO";       
    case SQL_NO_DATA_FOUND      : return "SQL_NO_DATA_FOUND" ;
	case SQL_STILL_EXECUTING    : return "SQL_STILL_EXECUTING (unsupported)"; 
	case SQL_NEED_DATA          : return "SQL_NEED_DATA (unsupported)";
    default                     : return "Unexpected sql retcode";
  } 
}

//---------------------------------------------------------------------------
// ODBC Environment class
//---------------------------------------------------------------------------  
  
ODBCEnvironment::ODBCEnvironment()        
                :m_hEnvironment(SQL_NULL_HENV)
{
}

ODBCEnvironment::~ODBCEnvironment()
{                                  
  if (Valid()) 
	  FreeEnv(); 
}  

BOOL ODBCEnvironment::AllocEnv()        
{            
  if (Valid()) return FALSE; // Already allocated

  m_Retcode=::SQLAllocEnv(&m_hEnvironment); 
  if (m_Retcode!=SQL_SUCCESS)
  { 
    ErrorInformation("SQLAllocEnv");
    m_hEnvironment=SQL_NULL_HENV;
    return FALSE;
  }
  else return TRUE;
}  
  
BOOL ODBCEnvironment::FreeEnv()
{                                  
  if (!Valid()) return FALSE; // Never allocated

  m_Retcode=::SQLFreeEnv(m_hEnvironment);
  if (m_Retcode!=SQL_SUCCESS)
  { 
    ErrorInformation("SQLFreeEnv");
    m_hEnvironment=SQL_NULL_HENV;
    return FALSE;
  }
  
  m_hEnvironment=SQL_NULL_HENV;
  return TRUE;
}                   
                   
//--------------------------------------------------------------------------------------------                   
// ODBC Connection class
//--------------------------------------------------------------------------------------------                                      

// Counter of connections, set to zero by compiler
int ODBCConnection::s_CountConnections;
                        
                        
ODBCConnection::ODBCConnection()
               :m_hConnection(SQL_NULL_HDBC),m_CountStatements(0),
			    m_Connected(FALSE),m_cEnvironment(GetStaticEnv())
{
  ++s_CountConnections;
}

ODBCConnection::~ODBCConnection()
{ 
  ASSERT(m_CountStatements==0); // Disconnecting with active statements?

  if (IsConnected()) 
      Disconnect();

  if (Valid()) 
     FreeConnect();

  if (!--s_CountConnections) 
     m_cEnvironment.FreeEnv(); // Release environment
}

// Static environment object inside this function
ODBCEnvironment & ODBCConnection::GetStaticEnv()
{
  static ODBCEnvironment Env;
  return Env;
}

BOOL ODBCConnection::AllocConnect()
{             
  // If already connected, return FALSE
  if (Valid()) return FALSE;
  // Allocate environment if not already done
  if (!m_cEnvironment.Valid() && !m_cEnvironment.AllocEnv()) return FALSE;

  // Create connection handle           
  m_Retcode=::SQLAllocConnect(m_cEnvironment.Handle(),&m_hConnection); 
  if (m_Retcode!=SQL_SUCCESS)
  { 
    ErrorInformation("SQLAllocConnect");
    m_hConnection=SQL_NULL_HDBC;
    return FALSE;
  }
  else
  {
    ASSERT(Valid()); 
    return TRUE;      
  }
}

BOOL ODBCConnection::FreeConnect()
{
  if (!Valid()) return FALSE; // Not connected  
  if (IsConnected()) Disconnect();

  m_Retcode=::SQLFreeConnect(m_hConnection);
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLFreeConnect");
  
  m_hConnection=SQL_NULL_HDBC;
  return Successful(m_Retcode);
}
  
BOOL ODBCConnection::Connect(LPCSTR ConnectionString,CString * pFinalString,HWND ParentWnd,BOOL DisableErrorHandling)
{             
  ASSERT(ConnectionString);
  ASSERT(ParentWnd==NULL || IsWindow(ParentWnd));
  
  // Check parameters                                          
  if (!ConnectionString) return FALSE;                                            
  // If handle not allocated, allocate it
  if (!Valid() && !AllocConnect()) return FALSE;
  // If we are already connected, disconnect
  if (IsConnected()) Disconnect();

  // Set hourglass cursor
  HCURSOR hCurOld=SetCursor(LoadCursor(NULL,IDC_WAIT));

  char  buff[2048];
  SWORD bufflen=0;           
  
  // Connect to database              
  // If the connecting string is empty, request login
  m_Retcode=::SQLDriverConnect(m_hConnection,ParentWnd,(UCHAR FAR *)ConnectionString,SQL_NTS,
                           (UCHAR FAR *)buff,sizeof(buff),&bufflen,
                           *ConnectionString?SQL_DRIVER_COMPLETE:SQL_DRIVER_PROMPT);
                    
  if (m_Retcode!=SQL_SUCCESS && m_Retcode!=SQL_SUCCESS_WITH_INFO)
  {                                             
    sprintf(buff,"SQLDriverConnect(%s)",(LPCSTR)ConnectionString);
    ErrorInformation(buff,DisableErrorHandling);
    SetCursor(hCurOld);
    return FALSE;
  }        

  // Store connection string                           
  ASSERT(bufflen<sizeof(buff));
  if (pFinalString) *pFinalString=buff;
  ConnectionString=NULL; // Cannot use the passed address anymore, in case
                         // it was changed above
  SetCursor(hCurOld);
  ASSERT(Valid());              
  return m_Connected=TRUE;                     
}  
  
BOOL ODBCConnection::Disconnect()
{ 
  if (!IsConnected()) return FALSE; // Not connected  

  m_Retcode=::SQLDisconnect(m_hConnection);
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLDisconnect");

  m_Connected=FALSE;
  return Successful(m_Retcode);
}  

BOOL ODBCConnection::SetConnectOption(UWORD fOption, UDWORD vParam)                                    
{
  if (!IsConnected()) return FALSE;
  m_Retcode=::SQLSetConnectOption(m_hConnection,fOption,vParam);
  
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLSetConnectOption(UDWORD)");
  return Successful(m_Retcode);
}

BOOL ODBCConnection::SetConnectOption(UWORD fOption, LPCSTR str)                                    
{ 
  ASSERT(str && lstrlen(str)<=256);
  if (!IsConnected()) return FALSE;
  
  m_Retcode=::SQLSetConnectOption(m_hConnection,fOption,(UDWORD)str);
  
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLSetConnectOption(LPCSTR)");
  return Successful(m_Retcode);
}

BOOL ODBCConnection::GetConnectOption(UWORD fOption, UDWORD * vParam)
{
  if (!IsConnected()) return FALSE;
  m_Retcode=::SQLGetConnectOption(m_hConnection,fOption,vParam);
  
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLGetConnectOption(UDWORD)");
  return Successful(m_Retcode);
}

BOOL ODBCConnection::GetConnectOption(UWORD fOption, LPSTR str, int size)
{
  ASSERT(size>256);
  if (!IsConnected() || size<=256) return FALSE;
  
  m_Retcode=::SQLGetConnectOption(m_hConnection,fOption,str);
  
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLGetConnectOption(LPSTR)");
  return Successful(m_Retcode);
}

BOOL ODBCConnection::Transact(UWORD fType)
{
  ASSERT(IsConnected() && m_cEnvironment.Valid());
  if (!IsConnected() || !m_cEnvironment.Valid()) return FALSE; 
  
  m_Retcode=::SQLTransact(m_cEnvironment.Handle(),m_hConnection,fType); 

  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLTransact");
  return Successful(m_Retcode);
}                                

BOOL ODBCConnection::GetInfo(UWORD fInfoType,LPSTR str, int size)
{
  ASSERT(str && size>0);
  if (!IsConnected()) return FALSE;
            
  m_Retcode=::SQLGetInfo(m_hConnection,fInfoType,str,size,NULL);
  
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLGetInfo");
  return Successful(m_Retcode);
}
                                
//-------------------------------------------------------------------------------------    
// ODBC Statement class
//-------------------------------------------------------------------------------------    

ODBCStatement::ODBCStatement(ODBCConnection * pCon, LPCSTR description)
              :m_hStatement(SQL_NULL_HSTMT),m_pConnection(pCon),
               m_NumErrors(0),m_NumWarnings(0),m_Description(description),
               m_CurrentCol(0),m_CurrentRow(0)
{ 
    VERIFY(pCon && AllocStmt(pCon));
}

ODBCStatement::ODBCStatement(LPCSTR description)
              :m_hStatement(SQL_NULL_HSTMT),m_pConnection(NULL),
               m_NumErrors(0),m_NumWarnings(0),m_Description(description),
               m_CurrentCol(0),m_CurrentRow(0)      
{ 
}

ODBCStatement::ODBCStatement(HSTMT hStmt, LPCSTR description)
              :m_hStatement(hStmt),m_pConnection(NULL),
               m_NumErrors(0),m_NumWarnings(0),m_Description(description),
               m_CurrentCol(0),m_CurrentRow(0)      
{
   ASSERT(Valid());
}


ODBCStatement::~ODBCStatement()
{
   if (Valid()) 
	   FreeStmt(SQL_DROP);
}

BOOL ODBCStatement::AllocStmt(ODBCConnection * pCon)
{ 
  ASSERT(!Valid() && pCon && pCon->Valid()); 
  if (Valid() || !pCon || !pCon->Valid()) return FALSE; // Can't connect if already connected
  
  if (!AllocStmt(pCon->Handle()))
     return FALSE;
  
  (m_pConnection=pCon)->AddStmt();
  return TRUE;
}
       
// This one does not take a ODBCConnection pointer but just an HDBC
BOOL ODBCStatement::AllocStmt(HDBC hCon)
{ 
  ASSERT(hCon!=SQL_NULL_HDBC && !Valid());                  
  if (Valid() || hCon==SQL_NULL_HDBC) return FALSE; // Can't connect if already connected
  
  m_Retcode=::SQLAllocStmt(hCon,&m_hStatement);
  if (m_Retcode!=SQL_SUCCESS)
  {
    ErrorInformation("SQLAllocStmt");
    m_hStatement=SQL_NULL_HSTMT;
    return FALSE;
  } 
                                  
  return TRUE;
}
       
BOOL ODBCStatement::FreeStmt(UWORD option)
{                                                
  if (!Valid()) return FALSE; // Not allocated

  m_Retcode=::SQLFreeStmt(m_hStatement,option);
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLFreeStmt");
  m_CurrentRow=m_CurrentCol=0;
 
  if (option==SQL_DROP)
  {
    if (m_pConnection) m_pConnection->DelStmt(); // Decrease counter of statements on this connection
    m_pConnection=NULL;
    m_hStatement=SQL_NULL_HSTMT;
    m_LastCommand.Empty();
    m_NumErrors=m_NumWarnings=0; // Reset number of errors
    return TRUE; 
  }   
  
  if (option==SQL_CLOSE) 
	  m_LastCommand.Empty();
  
  return Successful(m_Retcode);
}       

BOOL ODBCStatement::SetStmtOption(UWORD fOption, UDWORD vParam)                                    
{
  if (!Valid()) return FALSE;

  m_Retcode=::SQLSetStmtOption(m_hStatement,fOption,vParam);
  
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLSetStmtOption(UDWORD)");
  return Successful(m_Retcode);
}

BOOL ODBCStatement::SetStmtOption(UWORD fOption, LPCSTR str)                                    
{ 
  ASSERT(str && lstrlen(str)<=256);
  if (!Valid()) return FALSE;
  
  m_Retcode=::SQLSetStmtOption(m_hStatement,fOption,(UDWORD)str);
  
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLSetStmtOption(LCPSTR)");
  return Successful(m_Retcode);
}


BOOL ODBCStatement::GetStmtOption(UWORD fOption, UDWORD * vParam)
{
  if (!Valid()) return FALSE;
  m_Retcode=::SQLGetStmtOption(m_hStatement,fOption,vParam);
  
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLGetStmtOption(UDWORD)");
  return Successful(m_Retcode);
}

BOOL ODBCStatement::GetStmtOption(UWORD fOption, LPSTR str, int size)
{
  ASSERT(size>256);
  if (!Valid() || size<=256) return FALSE;
  
  m_Retcode=::SQLGetStmtOption(m_hStatement,fOption,str);
  
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLGetStmtOption(LPSTR)");
  return Successful(m_Retcode);
}

BOOL ODBCStatement::SetCursorName(LPCSTR name)          
{ 
  ASSERT(name && lstrlen(name)<=18);
  if (!Valid()) return FALSE;
                              
  m_Retcode=::SQLSetCursorName(m_hStatement,(UCHAR FAR *)name,lstrlen(name));
                              
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLSetCursorName");
  return Successful(m_Retcode);
}

BOOL ODBCStatement::GetCursorName(LPSTR name, int size)          
{ 
  ASSERT(name && size>18);
  if (!Valid()) return FALSE;

  SWORD actualLength;                            
  m_Retcode=::SQLGetCursorName(m_hStatement,(UCHAR FAR *)name,size,&actualLength);
                              
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLGetCursorName");
  return Successful(m_Retcode);
}

BOOL ODBCStatement::ExecDirect(LPCSTR SqlCommand, BOOL DisableErrorNotification)
{           
  ASSERT(Valid() && SqlCommand && *SqlCommand);  
  if (!Valid() || !SqlCommand || !*SqlCommand) return FALSE;

  m_LastCommand=SqlCommand;
  m_Retcode=::SQLExecDirect(m_hStatement,(UCHAR FAR *)(LPCSTR)m_LastCommand,SQL_NTS);
  m_CurrentRow=m_CurrentCol=0;
  
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLExecDirect",DisableErrorNotification);                              
  return m_Retcode==SQL_SUCCESS;
}

BOOL ODBCStatement::Prepare(LPCSTR SqlCommand, BOOL DisableErrorNotification)
{                  
  ASSERT(Valid() && SqlCommand && *SqlCommand);  
  if (!Valid() || !SqlCommand || !*SqlCommand) return FALSE;

  m_LastCommand=SqlCommand;
  m_Retcode=::SQLPrepare(m_hStatement,(UCHAR FAR *)(LPCSTR)m_LastCommand,SQL_NTS);  
  m_CurrentRow=m_CurrentCol=0;

  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLPrepare",DisableErrorNotification);                              
  return m_Retcode==SQL_SUCCESS;
}                 

BOOL ODBCStatement::Execute(BOOL DisableErrorNotification)
{                   
  ASSERT(Valid());
  if (!Valid()) return FALSE;
  
  m_Retcode=::SQLExecute(m_hStatement);
  m_CurrentRow=m_CurrentCol=0;
  
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLExecute",DisableErrorNotification);                              
  return m_Retcode==SQL_SUCCESS;
}

BOOL ODBCStatement::Fetch(BOOL WarningIfNoData, BOOL DisableErrorNotification)
{ 
  ASSERT(Valid());
  if (!Valid()) return FALSE; 
  
  m_Retcode=::SQLFetch(m_hStatement);
  m_CurrentRow++; m_CurrentCol=0; 
  
  if (m_Retcode==SQL_SUCCESS) return TRUE;                          
  if (m_Retcode==SQL_NO_DATA_FOUND && !WarningIfNoData) return FALSE;
                            
  ErrorInformation("SQLFetch",DisableErrorNotification);               
  return Successful(m_Retcode);  
}                       

BOOL ODBCStatement::SetPos(UWORD iRow, UWORD fOption, UWORD fLock)          
{ 
  ASSERT(Valid());
  if (!Valid()) return FALSE;

  m_Retcode=::SQLSetPos(m_hStatement,iRow,fOption,fLock);
                                                      
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLSetPos");
  return Successful(m_Retcode);
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

// Helper function used to retrieve data from selected columns after a successful fetch
ODBCStatement::EGetData ODBCStatement::GetDataHelper(int column, SWORD cType, LPVOID buff, long len, SDWORD * pcbValue)
{  
   ASSERT(Valid() && len>=0 && column>0);    
   ASSERT(buff || !len); // If you don't pass the buffer, you must set len to 0

   if (!Valid() || len<0 || column<=0) return GD_FALSE;

   SDWORD dummy=0;
   SDWORD & readlen = (pcbValue?*pcbValue:dummy);

   char   otherbuff[1]; // If user did not pass a buffer, make odbc happy
   LPVOID actualbuff = (buff?buff:otherbuff);

   m_Retcode=::SQLGetData(m_hStatement,m_CurrentCol=column,cType,actualbuff,len,&readlen);
           
   // These are the cases which are ok
   if (m_Retcode==SQL_SUCCESS && readlen==SQL_NULL_DATA) return GD_NULL_DATA; // Field is null 
   if (m_Retcode==SQL_SUCCESS                          ) return GD_TRUE; // All data has been retrieved      
   if (m_Retcode==SQL_SUCCESS_WITH_INFO                ) return GD_MORE_DATA; // More data to be retrieved       
   if (m_Retcode==SQL_NO_DATA_FOUND                    ) return GD_NO_MORE_DATA; // All data has been read already
   
   // If no data found then it's ok, just return FALSE
   ShowColumnInformation("SQLGetData",column);
   return GD_FALSE;
}    

// Helper function to bind the columns of a select statement to buffers that will receive the data
BOOL ODBCStatement::BindColHelper(int column,SWORD cType, LPVOID buff, long len, SDWORD * pcbValue)
{
   ASSERT(Valid() && buff && len>=0 && column>0);
   if (!Valid() || !buff) return FALSE;    

   static SDWORD DefaultUnusedValue;
   DefaultUnusedValue=0;

   if (!pcbValue) pcbValue=&DefaultUnusedValue; // SQLBindCol does not like receiving a NULL pcbValue
   m_Retcode=::SQLBindCol(m_hStatement,m_CurrentCol=column,cType,buff,len,pcbValue);
   
   if (m_Retcode!=SQL_SUCCESS) ShowColumnInformation("SQLBindCol",column);
   return Successful(m_Retcode);
}

// helper function used to set the value of the parameters in SQL statements
BOOL ODBCStatement::BindParameterHelper(int index, SWORD cType, SWORD sqlType, LPCVOID buff, long len,SDWORD * pcbValue)
{
   ASSERT(Valid() && index>0 && buff && len>=0);
   if (!Valid() || !buff) return FALSE;

   m_Retcode=::SQLBindParameter(m_hStatement,index,SQL_PARAM_INPUT,cType,sqlType,len,0,(LPVOID)buff,len,pcbValue);
   
   if (m_Retcode!=SQL_SUCCESS) ShowParamInformation("SQLBindParameter",index);
   return Successful(m_Retcode);
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

// This is a special function used to get long memo fields with multiple calls to SQLGetData
ODBCStatement::EGetData ODBCStatement::GetData(int column, CString * str, SDWORD * pcbValue)
{  
   ASSERT(Valid() && column>0 && str);    
   if (!Valid() || !str) return GD_FALSE;
     
   SDWORD dummy;
   SDWORD & readlen = (pcbValue?*pcbValue:dummy);
 
   char buff[4096];   *buff=0;
   str->Empty();
   
   // Read again and again if the data is long            
   while((m_Retcode=::SQLGetData(m_hStatement,m_CurrentCol=column,SQL_C_CHAR,buff,sizeof(buff), &readlen))==SQL_SUCCESS_WITH_INFO)
     (*str)+=buff, *buff=0;
   
   // Then append last part
   if (m_Retcode==SQL_SUCCESS) 
     (*str)+=buff;  

   // These are the cases which are ok
   if (m_Retcode==SQL_SUCCESS && readlen==SQL_NULL_DATA) return GD_NULL_DATA; // Field is null 
   if (m_Retcode==SQL_SUCCESS                          ) return GD_TRUE; // All data has been retrieved      
   if (m_Retcode==SQL_NO_DATA_FOUND                    ) return GD_NO_MORE_DATA; // All data has been read already
   
   // If no data found then it's ok, just return FALSE
   ShowColumnInformation("SQLGetData",column);
   return GD_FALSE;
}    

// Special for getting CTime
ODBCStatement::EGetData ODBCStatement::GetData(int column, CTime * time, SDWORD * pcbValue)
{ 
  ASSERT(time);	
  TIMESTAMP_STRUCT ts;

  EGetData ret=GetDataHelper(column, SQL_C_TIMESTAMP, &ts,0,pcbValue);

  if (ret==GD_TRUE)
	  return TIMESTAMP_STRUCT_to_CTime(ts,*time)?GD_TRUE:GD_FALSE;

  return ret;
}    

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------


BOOL ODBCStatement::RowCount(SDWORD * count)
{ 
  ASSERT(Valid() && count);    
  if (!Valid() || !count) return -1;

  m_Retcode=::SQLRowCount(m_hStatement,count);

  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLRowCount");
  return Successful(m_Retcode);
}

BOOL ODBCStatement::NumResultCols(SWORD * count)
{ 
  ASSERT(Valid() && count);   
  if (!Valid() || !count) return -1;

  m_Retcode=::SQLNumResultCols(m_hStatement,count);

  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLNumResultCols");
  return Successful(m_Retcode);
}

BOOL ODBCStatement::NumParams(SWORD * count)
{ 
  ASSERT(Valid() && count);   
  if (!Valid() || !count) return -1;

  m_Retcode=::SQLNumParams(m_hStatement,count);

  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLNumParams");
  return Successful(m_Retcode);
}


//-----------------------------------------------------------------------------

BOOL ODBCStatement::ColAttributes(int column,UWORD descType,SDWORD * value)
{ 
  ASSERT(Valid() && value);                             
  if (!Valid() || !value) return FALSE;
  
  const SDWORD invalid=0x7FFFFFFE; // Max signed long-1, just an invalid value to see if we retrieve something
  SDWORD getvalue=invalid;
  
  *value=0;
  m_Retcode=::SQLColAttributes(m_hStatement,column,descType,0,0,0,&getvalue);
  
  ASSERT(!Successful(m_Retcode) || getvalue!=invalid); // Check that we got a valid value or failed
  if (getvalue!=invalid) *value=getvalue;
  
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLColAttributes(numeric)");
  return Successful(m_Retcode) && getvalue!=invalid;
}

BOOL ODBCStatement::ColAttributes(int column,UWORD descType,LPSTR buff,int len)
{ 
  ASSERT(Valid() && buff && len>0);                             
  if (!Valid() || !buff || len<=0) return FALSE;
  
  const SWORD invalid=32766; // Max signed it - 1, an invalid value
  SWORD available=invalid;
  
  *buff=0;
  m_Retcode=::SQLColAttributes(m_hStatement,column,descType,buff,len,&available,0);
  
  // you must pass a buffer which is long enough!
  ASSERT(!Successful(m_Retcode) || available<len && available!=invalid); 
    
  if (m_Retcode!=SQL_SUCCESS) ErrorInformation("SQLColAttributes(string)");
  return Successful(m_Retcode) && available<len;
}

//-------------------------------------------------------------------------

int ODBCStatement::GetNumErrors()
{ 
   int num=m_NumErrors;
   m_NumErrors=0; // Calling it resets it to 0!
   return num;
}

int ODBCStatement::GetNumWarnings()
{ 
   int num=m_NumWarnings;
   m_NumWarnings=0; // Calling it resets it to 0!
   return num;
}

//-------------------------------------------------------------------------
                                                    
void ODBCStatement::ShowColumnInformation(LPCSTR function, int column)
{                           
   char msg[256];
   sprintf(msg,"%s in column %i",function,column);
   ErrorInformation(msg);
}                                                    
  
void ODBCStatement::ShowParamInformation(LPCSTR function, int index)
{ 
   char msg[256];
   sprintf(msg,"%s in parameter %i",function,index);
   ErrorInformation(msg);
}                                                    
                                                      
// Error handling routines ------------------------------------------------------------------
                                                                         
                                                                         
static char * ErrorNotFound="No Error or SQLError failed getting problem description";                    
                                                                         
                                                                         
BOOL ODBCEnvironment::ErrorInformation(LPCSTR function,BOOL disableTemporarily) 
{      
    if (!function) return FALSE;        
    ASSERT(m_Retcode!=SQL_SUCCESS);

    char    errordescription[SQL_MAX_MESSAGE_LENGTH];
    SWORD  pcnmsg=0;
    m_NativeError=0;
    *errordescription=0;
    *m_ErrorState=0;

    if (SQL_SUCCESS!=SQLError(m_hEnvironment,SQL_NULL_HDBC,SQL_NULL_HSTMT,(UCHAR FAR*)m_ErrorState,&m_NativeError,(UCHAR FAR*)errordescription,sizeof(errordescription),&pcnmsg))
       if (!*errordescription) strcpy(errordescription,ErrorNotFound);
    
    if (m_ErrorHandlingEnabled && !disableTemporarily)
       return ReportODBCProblem(!IsWarning(m_Retcode),function,errordescription,m_ErrorState,GetRetcodeDescription(m_Retcode),"","ODBCEnvironment",m_NativeError,0);
    else
       return FALSE;  
}      


BOOL ODBCConnection::ErrorInformation(LPCSTR function,BOOL disableTemporarily) 
{
    if (!function) return FALSE;            
    ASSERT(m_Retcode!=SQL_SUCCESS);

    char errordescription[SQL_MAX_MESSAGE_LENGTH];
    SWORD  pcnmsg=0;
    m_NativeError=0;
    *errordescription=0;
    *m_ErrorState=0;

    if (SQL_SUCCESS!=SQLError(SQL_NULL_HENV,m_hConnection,SQL_NULL_HSTMT,(UCHAR FAR*)m_ErrorState,&m_NativeError,(UCHAR FAR*)errordescription,sizeof(errordescription),&pcnmsg))
        if (!*errordescription) strcpy(errordescription,ErrorNotFound);
                             
    if (m_ErrorHandlingEnabled && !disableTemporarily)                             
        return ReportODBCProblem(!IsWarning(m_Retcode),function,errordescription,m_ErrorState,GetRetcodeDescription(m_Retcode),"","ODBCConnection",m_NativeError,0);
    else
        return FALSE;    
}        

BOOL ODBCStatement::ErrorInformation(LPCSTR function,BOOL disableTemporarily) 
{
    if (!function) return FALSE;            
    ASSERT(m_Retcode!=SQL_SUCCESS);

    char  errordescription[SQL_MAX_MESSAGE_LENGTH];
    SWORD pcnmsg=0;
    m_NativeError=0;
    *errordescription=0;
    *m_ErrorState=0;
    
    CString Function;
    CString Extra;   
           
    if (IsWarning(m_Retcode)) m_NumWarnings++;
    else                    m_NumErrors++; // Count the errors!

    if (SQL_SUCCESS!=SQLError(SQL_NULL_HENV,SQL_NULL_HDBC,m_hStatement,(UCHAR FAR*)m_ErrorState,&m_NativeError,(UCHAR FAR*)errordescription,sizeof(errordescription),&pcnmsg)) 
        if (!*errordescription) strcpy(errordescription,ErrorNotFound);    
    
    Function=(m_Description+":")+function; // Show description too!  
    AdditionalInformation(Extra);        // Other data
    
    if (m_ErrorHandlingEnabled && !disableTemporarily)
       return ReportODBCProblem(!IsWarning(m_Retcode),Function,errordescription,m_ErrorState,GetRetcodeDescription(m_Retcode),m_LastCommand,Extra,m_NativeError,m_CurrentRow);
    else
       return FALSE;   
}             


// Find out some more information about error  
BOOL ODBCStatement::AdditionalInformation(CString & str)
{ 
  SWORD  sword=0;
  SDWORD sdword=0;
  
  char buff[1024],piece[256];
  *buff=0;
  

  if (SQLRowCount(m_hStatement,&sdword)==SQL_SUCCESS)
  {
	 sprintf(piece," RowCount=%li ",sdword);
     strcat(buff,piece);
 }
  
  if (SQLNumResultCols(m_hStatement,&sword)==SQL_SUCCESS)
  {
	 sprintf(piece," NumResultCols=%i ",sword);
     strcat(buff,piece); 
  }
	 
  if (SQLNumParams(m_hStatement,&sword)==SQL_SUCCESS)
  {
     sprintf(piece," NumParams=%i ",sword);
     strcat(buff,piece); 
  }
               
  ASSERT(lstrlen(buff)<sizeof(buff));             
  if (*buff) 
     str+=buff; 
  return *buff!=0;
}


//----------------------------------------------------------------------------------------------------------------

// Convert from a CTime to a TIMESTAMP_STRUCT
BOOL CTime_to_TIMESTAMP_STRUCT(const CTime & time, TIMESTAMP_STRUCT & t)
{
  ASSERT((*(long*)&time)>=0);
  if ((*(long*)&time)>=0)
  {
	t.year     = time.GetYear();
	t.month    = time.GetMonth();
	t.day      = time.GetDay();
	t.hour     = time.GetHour();
	t.minute   = time.GetMinute();
	t.second   = time.GetSecond();
	t.fraction = 0; 
	return TRUE;
  }
  else return FALSE;
}
      
// Convert a TIMESTAMP_STRUCT to a CTime
BOOL TIMESTAMP_STRUCT_to_CTime(const TIMESTAMP_STRUCT & t, CTime & time)
{
  if (t.year>=1970 && t.year<2038)
     time=CTime(t.year,t.month,t.day,t.hour,t.minute,t.second);
  else
     return FALSE; 
  return TRUE;
}
      
                                     
//----------------------------------------------------------------------------------------------------------------


/* Example of a function that reports ODBC errors
// Handles warnings generated by our ODBC classes
BOOL ReportODBCProblem(BOOL isError,LPCSTR function,LPCSTR error,LPCSTR state,LPCSTR retcode,LPCSTR command,LPCSTR extra,long native,long row)
{ 
  ASSERT(function && error && state && retcode && command && extra);
  
  CString All;
  char msg[1024]; 
  
  sprintf(msg,"ODBC %s: Function %s returned %s with state %s native error %li %s",
           isError?"Error":"Warning",function,retcode,state,native,extra);
  ASSERT(lstrlen(msg)<sizeof(msg));         

  All=msg;
  All+="\r\n";
  All+=error;
  All+="\r\n";
  if (command && *command) 
     { All+=command; All+="\r\n"; }
  if (row)
     All+=bsprintf("Row %li\r\n",row);      
  ASSERT(lstrlen(msg)<sizeof(msg));              
  
  if (isError) 
     MessageBox(AfxGetMainWnd()->GetSafeHwnd(),All,"ODBC Error",MB_OK|MB_ICONSTOP); 
  else
     TRACE(All);
        
  return 0;
}
*/  