#ifndef ODBCGIO_H
#define ODBCGIO_H

#include <sqlext.h>


//******************************************************************************
//
// Database Classes by G. Bavestrelli         Copyright Techint S.p.A. 1995-1999
// Any feedback is welcome, you can contact me at giovanni.bavestrelli@pomini.it
//
// These classes encapsulate most of the ODBC 2.0 API to access databases.
// To use them you have to include ODBCGIO.H in your source files that access 
// the databases and include ODBCGIO.CPP in your project.
// The two classes you will use are ODBCConnection and ODBCStatement.
// Class ODBCEnvironment is transparent and you will not need to use it directly.
// Most of the member functions of these classes are named exaclty as the ODBC 
// functions they wrap but without the SQL prefix.
// For documentation, see the ODBC API help for the corresponding function
// (the function with the same name but with SQL at the beginning, if there is one).
// For the meaning of the parameter pcbValue (often used) you have to see the 
// ODBC documentation for the relative function.
// These classes implement a thin and hopefully efficent wrapper of the ODBC API.
// For the meaning and use of each function you should refer to the ODBC API
// and the source code ODBCGIO.CPP.
// 
// To use these classes you also have to define the function ReportODBCProblem.
// The prototype for this function is in ODBCGIO.CPP 
// Such function will receive all the error information in case of ODBC errors.
// Feel free to modify this error handling scheme, as it is the simplest posible.
// (It is also rather ugly, but it works and it is easy to improve it).
//
// Most of the functions in these classes are defined as BOOL.
// They return FALSE in case of errors, but in case of errors they also call
// the ReportODBCProblem function with the error description
// (unless you disable error handling, which you can do at any time).
//
// Note that these classes were designed to encapsulate the ODBC 2.0 API
// This API still works with the new ODBC 3.0, but I use the old API.
// Microsoft's ODBC classes that come with Visual C++ 6.0 also use ODBC 2.0
//
// These classes were designed to work only in MFC applications.
// Thread safety was NOT an issue for me when I designed these classes.
// So I make NO guarantee as to the thread safety of these classes.
// (acutally I better make no guarantee at all about anything these days)
// 
// These classes were not designed for derivation
// That is why the base class constructor is not virtual
//
// Only synchronous ODBC calls are supported by these classes.
// ODBC return types SQL_STILL_EXECUTING and SQL_NEED_DATA are NOT supported.
// (Microsoft's ODBC classes also don't support asynchronous ODBC calls).
//
// UNICODE is also NOT supported by these classes.
// These classes also don't throw nor catch exceptions.
//
// By now you probably think there is more stuff unsupported than supported.
// In this case, feel free to use other ODBC classes. 
// I tend to support what I need, also because there is a bigger chance to 
// be able to test what I need than what I think would be nice.
// So when I will need thread safety, UNICODE, SQL_STILL_EXECUTING etc.
// I will modify these classes to support them. Until then....
//
//*****************************************************************************
//
// Brief SQL syntax reminder :
//
// "SELECT [Column 1], [Column 2] FROM TableName WHERE Column3=? ORDER BY 1"
// "INSERT INTO [Table Name] ([Column 1],[Column 2],Column3) VALUES (?,?,?)"
// "UPDATE TableName SET Column1=?, [Column 2]=? WHERE [Column 3]=?"
// "DELETE FROM [Table Name] WHERE Column1=?" AND [Column 2]=2
//
// You can use any SQL command with these classes   
//
//*****************************************************************************
//
// ODBC Connection string syntax:  
//
//    "DSN=SourceDB;UID=UserId;PWD=YourPassword"
//
// None of these sections (DSN,UID,PWD) is compulsory, and more sections exist
// If you don't specify any (pass NULL) a message box will appear so you can
// select the database at run time.
//
//*****************************************************************************

// Backward compatibility with Win 3.1
#ifndef WIN32
typedef const void FAR * LPCVOID;
#endif

//=============================================================================

class ODBCEnvironment;
class ODBCConnection;
class ODBCStatement;           

//=============================================================================           

// Base class for all ODBC classes
class ODBCBase
{            
 protected:
 
  RETCODE m_Retcode;        // Last RETCODE returned from ODBC function
  long    m_NativeError;    // Last native error
  char    m_ErrorState[16]; // State for last error found (not reset if no error found!) 

  BOOL    m_ErrorHandlingEnabled;
 
  ODBCBase():m_Retcode(SQL_SUCCESS),m_ErrorHandlingEnabled(TRUE),m_NativeError(0)
       {*m_ErrorState=0;}
          
 public:

 ~ODBCBase(){}
	 
  RETCODE Retcode() const { return m_Retcode; }        
  LPCSTR  GetErrorState() const  { return m_ErrorState  ; }
  long    GetNativeError() const { return m_NativeError; }
  void    ClearLastError() { m_NativeError=0; *m_ErrorState=0; m_Retcode=SQL_SUCCESS; }
 
                          
  // Enable/disable calling the ReportODBCProblem function for each ODBC error
  void    EnableErrorHandling(BOOL enable) { m_ErrorHandlingEnabled=enable; }
  BOOL    IsErrorhandlingEnabled() const   { return m_ErrorHandlingEnabled; }

 private:

   operator = (const ODBCBase & d);  // Disable operator equal
   ODBCBase(const ODBCBase &);       // Disable copy constructor
};

//=============================================================================           

// Environment class is private and is only used by class ODBCConnection          
class ODBCEnvironment : public ODBCBase      
{
  HENV m_hEnvironment;

 protected:
 
  ODBCEnvironment();
  
 public: 
  
 ~ODBCEnvironment();
  
 protected:
 
  BOOL AllocEnv();
  BOOL FreeEnv();
  
  BOOL Valid()  const { return m_hEnvironment!=SQL_NULL_HENV; }
  HENV Handle() const { return m_hEnvironment; }

  BOOL ErrorInformation(LPCSTR msg,BOOL disableTemporarily=FALSE);
  
  friend class ODBCConnection;
};
          
//=============================================================================           
// Class that represents the connection to a database.
// You need one object of this class for each database you want to access.
//=============================================================================           
          
class ODBCConnection : public ODBCBase
{
  ODBCEnvironment & m_cEnvironment;
  static int s_CountConnections;

  HDBC m_hConnection;
  int  m_CountStatements;  
  BOOL m_Connected;

 public:
 
  ODBCConnection();
 ~ODBCConnection(); 

  BOOL AllocConnect();
  BOOL FreeConnect();

  // Connect to database passing ODBC connection string
  // The full connection string is returned in FinalConnectionString
  BOOL Connect(LPCSTR ConnectionString,CString * FinalConnectionString=NULL,HWND ParentWnd=NULL,BOOL DisableErrorHandlingWhileConnecting=FALSE);
  BOOL Disconnect(); 

  BOOL Valid()      const { return m_hConnection!=SQL_NULL_HDBC; }
  HDBC Handle()     const { return m_hConnection; }
  BOOL IsConnected()const { return m_hConnection!=SQL_NULL_HDBC && m_Connected; }   
                                       
  BOOL SetConnectOption(unsigned short fOption, unsigned long vParam);                                    
  BOOL SetConnectOption(unsigned short fOption, LPCSTR str);                                    
  BOOL GetConnectOption(unsigned short fOption, unsigned long * vParam);
  BOOL GetConnectOption(unsigned short fOption, LPSTR str, int size);
                                 
  BOOL GetInfo(unsigned short fInfoType,LPSTR str, int size);
  BOOL GetInfo(unsigned short fInfoType,short * num)
       { return GetInfo(fInfoType,(LPSTR)num,sizeof(num)); }
  BOOL GetInfo(unsigned short fInfoType,DWORD & num)          
       { return GetInfo(fInfoType,(LPSTR)num,sizeof(num)); }
                                                
  // Transactions                                              
  BOOL EnableTransactions(BOOL On)
       { return SetConnectOption(SQL_AUTOCOMMIT,On?SQL_AUTOCOMMIT_OFF:SQL_AUTOCOMMIT_ON);}
  BOOL Transact(unsigned short fType);     
  BOOL Commit()
       { return Transact(SQL_COMMIT);}
  BOOL Rollback()
       { return Transact(SQL_ROLLBACK);}
       
 private:                           
                            
  void AddStmt() { m_CountStatements++; ASSERT(m_CountStatements> 0);} 
  void DelStmt() { m_CountStatements--; ASSERT(m_CountStatements>=0);}

  static ODBCEnvironment & GetStaticEnv();

 protected:

  BOOL ErrorInformation(LPCSTR msg,BOOL disableTemporarily=FALSE);

  friend class ODBCStatement;
};                 

                   
//=============================================================================
// This class encapsulates the ODBC Statements
// Most of the work with the database will be done through one of these objects
//=============================================================================
    
class ODBCStatement : public ODBCBase   
{
  HSTMT m_hStatement;          // Statement handle 
  CString m_LastCommand,m_Description; // Information for errors          
  ODBCConnection * m_pConnection;    // Database connection object
  int m_NumErrors,m_NumWarnings; // Number of errors that have occurred
  int  m_CurrentCol;  // What column are we positioned on
  long m_CurrentRow;  // What row are we positioned on
  
 public:
  
  // When you create a statement you must pass it a ODBCConnection pointer or and HSTMT handle.
  // Description is just a text that will be appended to error messages.   
  ODBCStatement(ODBCConnection * pCon, LPCSTR description=NULL);
  ODBCStatement(HSTMT hstmt, LPCSTR description=NULL);  
  ODBCStatement(LPCSTR description=NULL);
 ~ODBCStatement(); 
  
  // Allocate statement if you didn't pass the necessary arguments to the constructor
  BOOL AllocStmt(ODBCConnection * pCon); 
  BOOL AllocStmt(HDBC hCon);         
  
  BOOL FreeStmt(unsigned short option=SQL_DROP); // It gets freed automatically by destructor!
  
  BOOL  Valid()  const   { return m_hStatement!=SQL_NULL_HSTMT; }
  HSTMT Handle() const   { return m_hStatement; }    
     
  BOOL SetStmtOption(unsigned short fOption, unsigned long vParam);                                    
  BOOL SetStmtOption(unsigned short fOption, LPCSTR str);                                    
  BOOL GetStmtOption(unsigned short fOption, unsigned long * vParam);
  BOOL GetStmtOption(unsigned short fOption, LPSTR str, int size);
                    
  BOOL SetCursorName(LPCSTR name);
  BOOL GetCursorName(LPSTR name, int size);                  
  
  // Main statement execution and fetching functions   
  // If you set DisableErrorNotification to TRUE, no error message will be processed in case of errors                                   
  BOOL ExecDirect(LPCSTR SqlCommand, BOOL DisableErrorNotification=FALSE);
  BOOL Execute(BOOL DisableErrorNotification=FALSE);
  BOOL Prepare(LPCSTR SqlCommand, BOOL DisableErrorNotification=FALSE);
  BOOL Fetch(BOOL GenerateWarningIfNoDataFound=FALSE, BOOL DisableErrorNotification=FALSE);
  
  BOOL SetPos(unsigned short irow, unsigned short fOption=SQL_POSITION, unsigned short fLock=SQL_LOCK_NO_CHANGE);

  // Get data from fields after a successful fetch operation
  // You can pass buff==NULL and size==0 to get just size of data
  enum EGetData { GD_FALSE, GD_TRUE, GD_NULL_DATA, GD_MORE_DATA, GD_NO_MORE_DATA };
  EGetData GetDataHelper(int column, short paramType, LPVOID buff, long size, long * pcbValue = NULL);// Helper
  EGetData GetDataBinary(int column,LPVOID buff,long size,long * pcbValue=NULL);  
  EGetData GetData(int column, LPSTR str ,   long size,long * pcbValue=NULL);
  EGetData GetData(int column, TIMESTAMP_STRUCT * time,long * pcbValue=NULL);
  EGetData GetData(int column, TIME_STRUCT      * time,long * pcbValue=NULL);
  EGetData GetData(int column, DATE_STRUCT      * date,long * pcbValue=NULL);
  EGetData GetData(int column, short  * num,           long * pcbValue=NULL);
  EGetData GetData(int column, long   * num,           long * pcbValue=NULL);
  EGetData GetData(int column, double * num,           long * pcbValue=NULL);
  EGetData GetData(int column, float  * num,           long * pcbValue=NULL);
  EGetData GetData(int column, int    * num,           long * pcbValue=NULL);
  EGetData GetData(int column, unsigned*num,           long * pcbValue=NULL);
  EGetData GetData(int column, CString *str,           long * pcbValue=NULL);// Reads long fields with multiple SQLGetData calls
  EGetData GetData(int column, CTime * time,           long * pcbValue=NULL);// Convert TIMESTAMP_STRUCT to CTime

  // Bind each column with a buffer that will receive the data after each fetch operation.
  // NB: The buffer must remain valid until all fetching has finished!!!
  BOOL BindColHelper(int column,short paramType, LPVOID buff, long len, long * pcbValue=NULL);// Helper
  BOOL BindColBinary(int column, LPVOID buff, long len,  long * pcbValue=NULL);  
  BOOL BindCol(int column, LPSTR buff,  long len,  long * pcbValue=NULL);
  BOOL BindCol(int column, TIMESTAMP_STRUCT * time,long * pcbValue=NULL);
  BOOL BindCol(int column, TIME_STRUCT      * time,long * pcbValue=NULL);
  BOOL BindCol(int column, DATE_STRUCT      * date,long * pcbValue=NULL);
  BOOL BindCol(int column, short  * num,           long * pcbValue=NULL);
  BOOL BindCol(int column, long   * num,           long * pcbValue=NULL);
  BOOL BindCol(int column, double * num,           long * pcbValue=NULL);
  BOOL BindCol(int column, float  * num,           long * pcbValue=NULL);
  BOOL BindCol(int column, int    * num,           long * pcbValue=NULL);
  BOOL BindCol(int column, unsigned*num,           long * pcbValue=NULL);
  

  // Bind each parameter with a buffer containing the value to use in place of the parameter
  // NB: The buffer must remain valid until you call execute!!!
  BOOL BindParameterHelper(int index, short paramType, short sqlType, LPCVOID buff, long len=0,long * pcbValue=NULL);// Helper
  BOOL BindParameterLongBinary(int index,LPCVOID buff,         long * pcbValue);     // Must specify length for binary
  BOOL BindParameterLongChar(int index,LPCSTR buff,            long * pcbValue=NULL);// Must specify length for long strings?
  BOOL BindParameterBinary(int index,LPCVOID buff,             long * pcbValue);     // Must specify length for binary
  BOOL BindParameter(int index,LPCSTR buff,                    long * pcbValue=NULL);
  BOOL BindParameter(int index,const TIMESTAMP_STRUCT * time,  long * pcbValue=NULL);
  BOOL BindParameter(int index,const TIME_STRUCT      * time,  long * pcbValue=NULL);
  BOOL BindParameter(int index,const DATE_STRUCT      * date,  long * pcbValue=NULL);
  BOOL BindParameter(int index,const short  * num,             long * pcbValue=NULL);
  BOOL BindParameter(int index,const long   * num,             long * pcbValue=NULL);
  BOOL BindParameter(int index,const double * num,             long * pcbValue=NULL);
  BOOL BindParameter(int index,const float  * num,             long * pcbValue=NULL);
  BOOL BindParameter(int index,const int    * num,             long * pcbValue=NULL);
  BOOL BindParameter(int index,const unsigned*num,             long * pcbValue=NULL);
  
  // Get information about result set
  BOOL RowCount(long * count);  // Number of rows affected by last statement     
  BOOL NumResultCols(short * count); // Number of columns in result set
  BOOL NumParams(short * count);// Number of parameters in SQL statement 

  // Get information about columns
  BOOL ColAttributes(int column,unsigned short descType,long * value);
  BOOL ColAttributes(int column,unsigned short descType,LPSTR buff,int len);
  
  BOOL GetColumnName(int column,LPSTR buff, int len);
  BOOL GetColumnType(int column,long * type);
                                                         
  int  GetNumErrors();   // Calling it resets counter to 0
  int  GetNumWarnings(); // Calling it resets counter to 0
  void ResetErrorCounters() { m_NumErrors=m_NumWarnings=0; }
           
  const CString & GetLastCommand() const { return m_LastCommand; }
  const CString & GetDescription() const { return m_Description; }

protected:
                                                                                              
  BOOL ErrorInformation(LPCSTR msg,BOOL disableTemporarily=FALSE);
  BOOL AdditionalInformation(CString & str);

  void ShowColumnInformation(LPCSTR message, int column);
  void ShowParamInformation(LPCSTR message, int index);
               
};

// NOTE: Above we use long  where below and in CPP file we use SDWORD
//       Above we use short where below and in CPP file we use SWORD
//       This is because I prefer to see simple long and short in interface
//       But I leave implementation correct so in case something changes
//       the compiler will warn me of the change
//       This is a questionable solution, I know, but I like it anyway.


//-----------------------------------------------------------------------------                       

inline BOOL ODBCStatement::GetColumnName(int column,LPSTR buff, int len)
{
  return ColAttributes(column,SQL_COLUMN_NAME,buff,len);
}

inline BOOL ODBCStatement::GetColumnType(int column,long * type)
{
  return ColAttributes(column,SQL_COLUMN_TYPE,type);
}

//-----------------------------------------------------------------------------


inline ODBCStatement::EGetData ODBCStatement::GetDataBinary(int column, LPVOID buf, long size, long * pcbValue)
{  
  return GetDataHelper(column, SQL_C_BINARY, buf, size, pcbValue);
}    

inline ODBCStatement::EGetData ODBCStatement::GetData(int column, LPSTR str, long size, long * pcbValue)
{  
  return GetDataHelper(column, SQL_C_CHAR, str, size, pcbValue);
}    

inline ODBCStatement::EGetData ODBCStatement::GetData(int column, TIMESTAMP_STRUCT * time, long * pcbValue)
{  
  return GetDataHelper(column, SQL_C_TIMESTAMP, time,0,pcbValue);
}    

inline ODBCStatement::EGetData ODBCStatement::GetData(int column, TIME_STRUCT * time, long * pcbValue)
{  
  return GetDataHelper(column, SQL_C_TIME, time,0,pcbValue);
}    

inline ODBCStatement::EGetData ODBCStatement::GetData(int column, DATE_STRUCT * date, long * pcbValue)
{  
  return GetDataHelper(column, SQL_C_DATE, date,0,pcbValue);
}    

inline ODBCStatement::EGetData ODBCStatement::GetData(int column, short * num, long * pcbValue)
{  
  return GetDataHelper(column, SQL_C_SSHORT, num,0,pcbValue);
}    

inline ODBCStatement::EGetData ODBCStatement::GetData(int column, long * num, long * pcbValue)
{  
  return GetDataHelper(column, SQL_C_SLONG, num,0,pcbValue);
}    

inline ODBCStatement::EGetData ODBCStatement::GetData(int column, double * num, long * pcbValue)
{ 
  return GetDataHelper(column, SQL_C_DOUBLE, num,0,pcbValue); 
}    

inline ODBCStatement::EGetData ODBCStatement::GetData(int column, float * num, long * pcbValue)
{ 
  return GetDataHelper(column, SQL_C_FLOAT, num,0,pcbValue); 
}    

inline ODBCStatement::EGetData ODBCStatement::GetData(int column, int * num, long * pcbValue)
{                              
#ifdef WIN32
  return GetDataHelper(column, SQL_C_SLONG,  num,0,pcbValue); 
#else                                                                      
  return GetDataHelper(column, SQL_C_SSHORT, num,0,pcbValue); 
#endif  
}    

inline ODBCStatement::EGetData ODBCStatement::GetData(int column, unsigned * num, long * pcbValue)
{                              
#ifdef WIN32
  return GetDataHelper(column, SQL_C_ULONG,  num,0,pcbValue); 
#else                                                                      
  return GetDataHelper(column, SQL_C_USHORT, num,0,pcbValue); 
#endif  
}    

//-----------------------------------------------------------------------------               

inline BOOL ODBCStatement::BindColBinary(int column,LPVOID buff, long len, long * pcbValue)
{
   return BindColHelper(column, SQL_C_BINARY, buff, len, pcbValue);
}

inline BOOL ODBCStatement::BindCol(int column,LPSTR buff, long len, long * pcbValue)
{
   return BindColHelper(column, SQL_C_CHAR, buff, len, pcbValue);
}

inline BOOL ODBCStatement::BindCol(int column,TIMESTAMP_STRUCT * time,long * pcbValue)
{
   return BindColHelper(column, SQL_C_TIMESTAMP, time, 0, pcbValue);
}

inline BOOL ODBCStatement::BindCol(int column,TIME_STRUCT * time,long * pcbValue)
{
   return BindColHelper(column, SQL_C_TIME, time, 0, pcbValue);
}

inline BOOL ODBCStatement::BindCol(int column,DATE_STRUCT * date,long * pcbValue)
{
   return BindColHelper(column, SQL_C_DATE, date, 0, pcbValue);
}

inline BOOL ODBCStatement::BindCol(int column,short * num,long * pcbValue)
{
   return BindColHelper(column, SQL_C_SSHORT, num, 0, pcbValue);
}

inline BOOL ODBCStatement::BindCol(int column,long * num,long * pcbValue)
{
   return BindColHelper(column, SQL_C_SLONG, num, 0, pcbValue);
}

inline BOOL ODBCStatement::BindCol(int column,double * num,long * pcbValue)
{
   return BindColHelper(column, SQL_C_DOUBLE, num, 0, pcbValue);
}

inline BOOL ODBCStatement::BindCol(int column,float * num,long * pcbValue)
{
   return BindColHelper(column, SQL_C_FLOAT, num, 0, pcbValue);
}

inline BOOL ODBCStatement::BindCol(int column,int * num, long * pcbValue)
{           
#ifdef WIN32
   return BindColHelper(column, SQL_C_SLONG,  num, 0, pcbValue);
#else   
   return BindColHelper(column, SQL_C_SSHORT, num, 0, pcbValue);   
#endif     
}

inline BOOL ODBCStatement::BindCol(int column,unsigned * num, long * pcbValue)
{           
#ifdef WIN32
   return BindColHelper(column, SQL_C_ULONG,  num, 0, pcbValue);
#else   
   return BindColHelper(column, SQL_C_USHORT, num, 0, pcbValue);   
#endif     
}


//-----------------------------------------------------------------------------

inline BOOL ODBCStatement::BindParameterLongBinary(int index,LPCVOID buff,long * pcbValue)
{                                                                           
   ASSERT(pcbValue);                                   
   long   cbColDef=(pcbValue && *pcbValue>0?*pcbValue:0);
   return BindParameterHelper(index,SQL_C_BINARY,SQL_LONGVARBINARY,buff,cbColDef,pcbValue);
}

inline BOOL ODBCStatement::BindParameterLongChar(int index,LPCSTR buff,long * pcbValue)
{                                                        
   long   cbColDef=(pcbValue && *pcbValue>0?*pcbValue:0);
   return BindParameterHelper(index,SQL_C_CHAR,SQL_LONGVARCHAR,buff,cbColDef,pcbValue);
}

inline BOOL ODBCStatement::BindParameterBinary(int index,LPCVOID buff,long * pcbValue)
{                                                              
   ASSERT(pcbValue);                   
   long   cbColDef=(pcbValue && *pcbValue>0?*pcbValue:0);
   return BindParameterHelper(index,SQL_C_BINARY,SQL_BINARY,buff,cbColDef,pcbValue);
}

inline BOOL ODBCStatement::BindParameter(int index,LPCSTR buff,long * pcbValue)
{                                   
   ASSERT(buff && lstrlen(buff)<256);
   return BindParameterHelper(index,SQL_C_CHAR,SQL_CHAR,buff,0,pcbValue);
}

inline BOOL ODBCStatement::BindParameter(int index,const TIMESTAMP_STRUCT * time,long * pcbValue)
{
   return BindParameterHelper(index,SQL_C_TIMESTAMP,SQL_TIMESTAMP,time,0,pcbValue);
}

inline BOOL ODBCStatement::BindParameter(int index,const TIME_STRUCT * time,long * pcbValue)
{
   return BindParameterHelper(index,SQL_C_TIME,SQL_TIME,time,0,pcbValue);
}

inline BOOL ODBCStatement::BindParameter(int index,const DATE_STRUCT * date,long * pcbValue)
{
   return BindParameterHelper(index,SQL_C_DATE,SQL_DATE,date,0,pcbValue);
}

inline BOOL ODBCStatement::BindParameter(int index,const short * num,long * pcbValue)
{
   return BindParameterHelper(index,SQL_C_SSHORT,SQL_SMALLINT,num,0,pcbValue);
}

inline BOOL ODBCStatement::BindParameter(int index,const long * num,long * pcbValue)
{
   return BindParameterHelper(index,SQL_C_SLONG,SQL_INTEGER,num,0,pcbValue);
}

inline BOOL ODBCStatement::BindParameter(int index,const double * num,long * pcbValue)
{
   return BindParameterHelper(index,SQL_C_DOUBLE,SQL_DOUBLE,num,0,pcbValue);
}

inline BOOL ODBCStatement::BindParameter(int index,const float * num,long * pcbValue)
{
   return BindParameterHelper(index,SQL_C_FLOAT,SQL_REAL,num,0,pcbValue);
}

inline BOOL ODBCStatement::BindParameter(int index,const int * num,long * pcbValue)
{           
#ifdef WIN32
   return BindParameterHelper(index,SQL_C_SLONG, SQL_INTEGER,num,0,pcbValue);
#else   
   return BindParameterHelper(index,SQL_C_SSHORT,SQL_INTEGER,num,0,pcbValue);   
#endif     
}

inline BOOL ODBCStatement::BindParameter(int index,const unsigned * num,long * pcbValue)
{           
#ifdef WIN32
   return BindParameterHelper(index,SQL_C_ULONG, SQL_INTEGER,num,0,pcbValue);
#else   
   return BindParameterHelper(index,SQL_C_USHORT,SQL_INTEGER,num,0,pcbValue);   
#endif     
}

//-----------------------------------------------------------------------------
// DateTime conversion functions
//-----------------------------------------------------------------------------

// Convert a CTime to a TIMSTAMP_STRUCT, if possible
BOOL CTime_to_TIMESTAMP_STRUCT(const CTime & time, TIMESTAMP_STRUCT & t);
// Convert a TIMESTAMP_STRUCT to a CTime, if possible
BOOL TIMESTAMP_STRUCT_to_CTime(const TIMESTAMP_STRUCT & t, CTime & time);


#endif


